@@ -7,6 +7,7 @@
 import json
 from collections import defaultdict
 from pathlib import Path
+from typing import Any, List, Tuple, Union
 
 from . import classes as c
 from . import language
@@ -38,12 +39,59 @@ def cached(funct):
     return cached_function
 
 
+def object_pairs_hook_defer_duplicate_keys(object_pairs: List[Tuple[str, Any]]):
+    keys = [x[0] for x in object_pairs]
+    try:
+        assert len(set(keys)) == len(keys)
+    except AssertionError:
+        raise AssertionError(
+            "Dictionary has entries with same name: {}".format(
+                [object_pairs[i] for i, k in enumerate(keys) if keys.count(k) > 1]
+            )
+        )
+    return dict(object_pairs)
+
+
 ###############################################################################
 @cached
 def _get_load(lang="en_US"):
     return language.get("load", lang)
 
 
+GENERAL_UNITS_PATH = TOPDIR.joinpath("units.json")
+GENERAL_ENTITIES_PATH = TOPDIR.joinpath("entities.json")
+
+
+def LANGUAGE_ENTITIES_PATH(lang="en_US"):
+    return TOPDIR.joinpath(language.topdir(lang), "entities.json")
+
+
+def LANGUAGE_UNITS_PATH(lang="en_US"):
+    return TOPDIR.joinpath(language.topdir(lang), "units.json")
+
+
+def _load_json(path_or_string: Union[Path, str]):
+    if isinstance(path_or_string, Path):
+        with path_or_string.open("r", encoding="utf-8") as jsonfile:
+            return jsonfile.read()
+    elif isinstance(path_or_string, str) and path_or_string.endswith(".json"):
+        with open(path_or_string, "r", encoding="utf-8") as jsonfile:
+            return jsonfile.read()
+    return path_or_string
+
+
+def _load_json_dict(path_or_string: Union[Path, str, dict]):
+    if isinstance(path_or_string, dict):
+        return path_or_string
+    return json.loads(
+        _load_json(path_or_string),
+        object_pairs_hook=object_pairs_hook_defer_duplicate_keys,
+    )
+
+
+CUSTOM_ENTITIES = defaultdict(dict)
+CUSTOM_UNITS = defaultdict(dict)
+
 ###############################################################################
 def to_int_iff_int(value):
     """
@@ -115,39 +163,29 @@ def get_key_from_dimensions(derived):
 
 ###############################################################################
 class Entities(object):
-    def __init__(self, lang="en_US"):
+    def __init__(self, entity_dicts: List[Union[Path, str, dict]]):
         """
         Load entities from JSON file.
         """
 
-        path = TOPDIR.joinpath("entities.json")
-        with path.open(encoding="utf-8") as file:
-            general_entities = json.load(file)
-        names = [i["name"] for i in general_entities]
-
-        try:
-            assert len(set(names)) == len(general_entities)
-        except AssertionError:  # pragma: no cover
-            raise Exception(
-                "Entities with same name: %s" % [i for i in names if names.count(i) > 1]
-            )
+        # Merge entity dictionarys
+        all_entities = defaultdict(dict)
+        for ed in entity_dicts:
+            for new_name, new_ent in _load_json_dict(ed).items():
+                all_entities[new_name].update(new_ent)
 
         self.names = dict(
             (
-                k["name"],
-                c.Entity(name=k["name"], dimensions=k["dimensions"], uri=k["URI"]),
+                name,
+                c.Entity(
+                    name=name,
+                    dimensions=props.get("dimensions", []),
+                    uri=props.get("URI"),
+                ),
             )
-            for k in general_entities
+            for name, props in all_entities.items()
         )
 
-        # Update with language specific URI
-        with TOPDIR.joinpath(language.topdir(lang), "entities.json").open(
-            "r", encoding="utf-8"
-        ) as file:
-            lang_entities = json.load(file)
-        for ent in lang_entities:
-            general_entities[ent["name"]].uri = ent["URI"]
-
         # Generate derived units
         derived_ent = defaultdict(set)
         for entity in self.names.values():
@@ -193,7 +231,9 @@ def entities(lang="en_US"):
     """
     Cached entity object
     """
-    return Entities(lang)
+    return Entities(
+        [GENERAL_ENTITIES_PATH, LANGUAGE_ENTITIES_PATH(lang), CUSTOM_ENTITIES]
+    )
 
 
 ###############################################################################
@@ -222,40 +262,26 @@ def get_derived_units(names):
 
 ###############################################################################
 class Units(object):
-    def __init__(self, lang="en_US"):
+    def __init__(self, unit_dict_json: List[Union[str, Path, dict]], lang="en_US"):
         """
         Load units from JSON file.
         """
-        self.lang = lang
 
         # names of all units
         self.names = {}
         self.symbols, self.symbols_lower = defaultdict(set), defaultdict(set)
         self.surfaces, self.surfaces_lower = defaultdict(set), defaultdict(set)
         self.prefix_symbols = defaultdict(set)
+        self.lang = lang
 
-        # Load general units
-        path = TOPDIR.joinpath("units.json")
-        with path.open(encoding="utf-8") as file:
-            general_units = json.load(file)
-        # load language specifics
-        path = TOPDIR.joinpath(language.topdir(lang), "units.json")
-        with path.open(encoding="utf-8") as file:
-            lang_units = json.load(file)
+        unit_dict = defaultdict(dict)
+        for ud in unit_dict_json:
+            for name, unit in _load_json_dict(ud).items():
+                for nname, nunit in self.prefixed_units(name, unit):
+                    unit_dict[nname].update(nunit)
 
-        unit_dict = {}
-        for unit in general_units.copy():
-            general_units.extend(self.prefixed_units(unit))
-        for unit in general_units:
-            unit_dict[unit["name"]] = unit
-        for unit in lang_units.copy():
-            lang_units.extend(self.prefixed_units(unit))
-        for unit in lang_units:
-            unit_dict[unit["name"]] = unit_dict.get(unit["name"], unit)
-            unit_dict[unit["name"]].update(unit)
-
-        for unit in unit_dict.values():
-            self.load_unit(unit)
+        for name, unit in unit_dict.items():
+            self.load_unit(name, unit)
 
         self.derived = get_derived_units(self.names)
 
@@ -267,25 +293,25 @@ class Units(object):
         self.surfaces_all = self.surfaces.copy()
         self.surfaces_all.update(self.surfaces_lower)
 
-    def load_unit(self, unit):
+    def load_unit(self, name, unit):
         try:
-            assert unit["name"] not in self.names
+            assert name not in self.names
         except AssertionError:  # pragma: no cover
-            msg = "Two units with same name in units.json: %s" % unit["name"]
+            msg = "Two units with same name in units.json: %s" % name
             raise Exception(msg)
 
         obj = c.Unit(
-            name=unit["name"],
+            name=name,
             surfaces=unit.get("surfaces", []),
             entity=entities().names[unit["entity"]],
-            uri=unit["URI"],
+            uri=unit.get("URI"),
             symbols=unit.get("symbols", []),
             dimensions=unit.get("dimensions", []),
             currency_code=unit.get("currency_code"),
             lang=self.lang,
         )
 
-        self.names[unit["name"]] = obj
+        self.names[name] = obj
 
         for symbol in unit.get("symbols", []):
             self.symbols[symbol].add(obj)
@@ -301,35 +327,28 @@ class Units(object):
             self.surfaces_lower[plural.lower()].add(obj)
 
     @staticmethod
-    def prefixed_units(unit):
-        prefixed = []
+    def prefixed_units(name, unit):
+        yield name, unit
         # If SI-prefixes are given, add them
         for prefix in unit.get("prefixes", []):
             assert (
                 prefix in METRIC_PREFIXES
-            ), "Given prefix '{}' for unit '{}' not supported".format(
-                prefix, unit["name"]
-            )
+            ), "Given prefix '{}' for unit '{}' not supported".format(prefix, name)
             assert (
                 len(unit["dimensions"]) <= 1
-            ), "Prefixing not supported for multiple dimensions in {}".format(
-                unit["name"]
-            )
+            ), "Prefixing not supported for multiple dimensions in {}".format(name)
 
             uri = METRIC_PREFIXES[prefix].capitalize() + unit["URI"].lower()
             # we usually do not want the "_(unit)" postfix for prefixed units
             uri = uri.replace("_(unit)", "")
 
-            prefixed_unit = {
-                "name": METRIC_PREFIXES[prefix] + unit["name"],
+            yield METRIC_PREFIXES[prefix] + name, {
                 "surfaces": [METRIC_PREFIXES[prefix] + i for i in unit["surfaces"]],
                 "entity": unit["entity"],
                 "URI": uri,
                 "dimensions": [],
                 "symbols": [prefix + i for i in unit["symbols"]],
             }
-            prefixed.append(prefixed_unit)
-        return prefixed
 
 
 @cached
@@ -337,7 +356,7 @@ def units(lang="en_US"):
     """
     Cached unit object
     """
-    return Units(lang)
+    return Units([GENERAL_UNITS_PATH, LANGUAGE_UNITS_PATH(lang), CUSTOM_UNITS], lang)
 
 
 ###############################################################################
@@ -352,3 +371,48 @@ def training_set(lang="en_US"):
                 training_set_ += json.load(train_file)
 
     return training_set_
+
+
+###############################################################################
+def add_custom_unit(name: str, **kwargs):
+    """
+    Adds a custom unit to the set of default units
+    Note: causes a reload of all units
+    :param name: Name of the unit to add, should preferably be unique,
+    otherwise will overwrite attributes in existing units
+    :param kwargs: properties of the unit as found in units.json, i.e. surfaces=["centimetre"]
+    """
+    CUSTOM_UNITS[name].update(kwargs)
+    _CACHE_DICT.clear()
+
+
+def remove_custom_unit(name: str):
+    """
+    Removes a unit from the set of custom units
+    Note: causes a reload of all units
+    :param name: Name of the unit to remove. This will not affect units that are loaded per default.
+    """
+    CUSTOM_UNITS.pop(name)
+    _CACHE_DICT.clear()
+
+
+def add_custom_entity(name: str, **kwargs):
+    """
+    Adds a custom entity to the set of default entities
+    Note: causes a reload of all entities
+    :param name: Name of the entity to add, should preferably be unique,
+    otherwise will overwrite attributes in existing entities
+    :param kwargs: properties of the entity as found in entities.json, i.e. surfaces=["centimetre"]
+    """
+    CUSTOM_ENTITIES[name].update(kwargs)
+    _CACHE_DICT.clear()
+
+
+def remove_custom_entity(name: str):
+    """
+    Removes a entity from the set of custom entities
+    Note: causes a reload of all entities
+    :param name: Name of the entity to remove. This will not affect entities that are loaded per default.
+    """
+    CUSTOM_ENTITIES.pop(name)
+    _CACHE_DICT.clear()
